package entity

import "fmt"
import "crypto/md5"
import "encoding/base64"
import "github.com/google/uuid"

// A hash represents a hash sum that fits within a uint64.
type Hash [8]byte

// NewHash creates a new truncated md5 hash using the specified data.
func NewHash (data []byte) Hash {
	sum := md5.Sum(data)
	hash := Hash { }
	copy(hash[:], sum[:8])
	return hash
}

// Number converts the hash into a uint64.
func (hash Hash) Number () uint64 {
	var number uint64
	for _, part := range hash {
		number <<= 8
		number |=  uint64(part)
	}
	return number
}

// Key globally indexes top level entities in contexts where modules matter.
type Key struct {
	Unit uuid.UUID

	// TypeName is the name of the type
	Type string

	// Constant is the name of the constant
	Constant string

	// Function is the name of the method or function
	Function string
}

func (key Key) String () string {
	return fmt.Sprintf("%v::%s", key.Unit, key.name())
}

// LinkName returns the name that the entity it refers to will be given when
// compiled.
func (key Key) LinkName () string {
	data := [16]byte(key.Unit)
	unit := base64.StdEncoding.EncodeToString(data[:])
	return fmt.Sprintf("%v::%s", unit, key.name())
}

func (key Key) name () string {
	out := ""
	switch {
	case key.Constant != "": out = key.Constant
	case key.Function != "": out = fmt.Sprintf("[%s]", key.Function)
	}
	
	if key.Type != "" {
		if out == "" {
			out = key.Type
		} else {
			out = fmt.Sprintf("%s.%s", key.Type, out)
		}
	}
	
	return out
}

// Hash returns a representation of the hash of this key that fits within a
// uint64.
func (key Key) Hash () Hash {
	return NewHash([]byte("Key:" + key.String()))
}

// OnlyType returns a copy of the key that refers to a type instead of a method
// or constant.
func (key Key) OnlyType () Key {
	key.Constant = ""
	key.Function = ""
	return key
}
