package generator

import "fmt"
import "errors"
import "git.tebibyte.media/fspl/fspl/llvm"
import "git.tebibyte.media/fspl/fspl/entity"
import "git.tebibyte.media/fspl/fspl/analyzer"

func (this *generator) generateEvaluatedInt (literal *entity.LiteralInt) (llvm.Const, error) {
	irType, err := this.generateType(analyzer.ReduceToBase(literal.Type()))
	if err != nil { return nil, err }

	base := analyzer.ReduceToBase(literal.Type())
	switch base.(type) {
	case *entity.TypeInt, *entity.TypeWord:
		return llvm.NewConstInt(irType.(*llvm.TypeInt), int64(literal.Value)), nil
	case *entity.TypeFloat:
		return llvm.NewConstFloat(irType.(*llvm.TypeFloat), float64(literal.Value)), nil
	default:
		return nil, errors.New(fmt.Sprintln("constant int can't be used as", base))
	}
}

func (this *generator) generateEvaluatedFloat (literal *entity.LiteralFloat) (llvm.Const, error) {
	irType, err := this.generateType(analyzer.ReduceToBase(literal.Type()))
	if err != nil { return nil, err }
	return llvm.NewConstFloat(irType.(*llvm.TypeFloat), literal.Value), nil
}

func (this *generator) generateEvaluatedString (literal *entity.LiteralString) (llvm.Const, error) {
	// TODO
	panic("BUG: not supported yet")
}

func (this *generator) generateEvaluatedArray (literal *entity.LiteralArray) (llvm.Const, error) {
	baseAny := analyzer.ReduceToBase(literal.Type())
	base, ok := baseAny.(*entity.TypeArray)
	if !ok { return nil, errors.New(fmt.Sprintln("constant array can't be used as", baseAny)) }
	
	irDestTypeAny, err := this.generateType(base)
	if err != nil { return nil, err }
	irDestType := irDestTypeAny.(*llvm.TypeArray)
	irElementType, err := this.generateType(base.Element)
	if err != nil { return nil, err }

	irArray := &llvm.ConstArray {
		Ty:       irDestType,
		Elements: make([]llvm.Const, base.Length),
	}
	for index := range irArray.Elements {
		if index < len(literal.Elements) {
			irElement, err := this.generateEvaluated(literal.Elements[index])
			if err != nil { return nil, err }
			irArray.Elements[index] = irElement
		} else {
			irArray.Elements[index] = llvm.NewConstZeroInitializer(irElementType)
		}
	}

	return irArray, nil
}

func (this *generator) generateEvaluatedStruct (literal *entity.LiteralStruct) (llvm.Const, error) {
	baseAny := analyzer.ReduceToBase(literal.Type())
	base, ok := baseAny.(*entity.TypeStruct)
	if !ok { return nil, errors.New(fmt.Sprintln("constant struct can't be used as", baseAny)) }
	
	irDestTypeAny, err := this.generateType(base)
	if err != nil { return nil, err }
	irDestType := irDestTypeAny.(*llvm.TypeStruct)
	
	irStruct := &llvm.ConstStruct {
		Ty:     irDestType,
		Fields: make([]llvm.Const, len(base.Members)),
	}
	for index, member := range base.Members {
		if pair, ok := literal.MemberMap[member.Name]; ok {
			irMember, err := this.generateEvaluated(pair.Value)
			if err != nil { return nil, err }
			irStruct.Fields[index] = irMember
		} else {
			irMemberType, err := this.generateType(member.Type())
			if err != nil { return nil, err }
			irStruct.Fields[index] = llvm.NewConstZeroInitializer(irMemberType)
		}
	}

	return irStruct, nil
}

func (this *generator) generateEvaluatedBoolean (literal *entity.LiteralBoolean) (llvm.Const, error) {
	return llvm.NewConstBool(bool(literal.Value)), nil
}

func (this *generator) generateEvaluatedNil (literal *entity.LiteralNil) (llvm.Const, error) {
	ty, err := this.generateType(literal.Type())
	if err != nil { return nil, err }
	return llvm.NewConstZeroInitializer(ty), nil
}
